declare var window: any;

/////////////////////////////////////////////////

export class Sound {
  public buffer: AudioBuffer;

  public loop: boolean;
  public isInContext = false;
  // True if the sound is currently playing
  public isPlaying = false;
  // public isMusic = false;

  constructor(buffer: AudioBuffer) {
    this.buffer = buffer;
    this.loop = false;
  }

  private source: AudioBufferSourceNode | undefined;
  private volume: GainNode | undefined;
  private panner: PannerNode | undefined;

  setContext(context: AudioContext) {
    this.volume = context.createGain();
    this.panner = context.createPanner();
    this.source = context.createBufferSource();

    this.source.onended = () => {
      this.isPlaying = false;
    }

    this.source.connect(this.volume);
    this.volume.connect(this.panner);

    this.source.buffer = this.buffer;
    this.source.loop = this.loop;
    this.isInContext = true;

    this.panner.panningModel = 'HRTF';
    this.panner.distanceModel = 'inverse';
    this.panner.refDistance = 1;
    this.panner.maxDistance = 1000;
    this.panner.rolloffFactor = 1;
    this.panner.coneInnerAngle = 360;
    this.panner.coneOuterAngle = 0;
    this.panner.coneOuterGain = 0;
  }

  get node(): AudioNode | undefined {
    // return (this.isMusic) ? this.volume : this.panner;
    return this.volume;
  }

  play(time: number = 0) {
    if (!this.isInContext) {
      throw new Error('Sound not in context');
    }
    if (this.source)
      this.source.start(time);
  }

  stop() {
    if (this.source)
      this.source.stop();
  }
}

/////////////////////////////////////////////////
export class SoundSystem {
  static addSoundToContext(context: AudioContext, sound: Sound) {
    sound.setContext(context);
  }
  static soundContext: AudioContext;
  static masterGainNode: GainNode;
  static analyser: AnalyserNode;

  constructor() {
    this.reboot();
  }

  reboot() {
    try {
      const AudioContext = window.webkitAudioContext // Safari and old versions of Chrome
        || window.AudioContext                       // Default
        || false;

      if (AudioContext) {
        if (SoundSystem.soundContext) {
          SoundSystem.soundContext.suspend();
        }
        SoundSystem.soundContext = new AudioContext();
        SoundSystem.soundContext.onstatechange = this.handleStateChange;

        SoundSystem.analyser = SoundSystem.soundContext.createAnalyser();

        SoundSystem.masterGainNode = SoundSystem.soundContext.createGain();
        SoundSystem.masterGainNode.connect(SoundSystem.soundContext.destination);

        SoundSystem.masterGainNode.connect(SoundSystem.analyser);

        window.addEventListener('focus', (event: any) => {
          // console.log('Andy: Resume Play');
          this.resumePlay();
        });
      } else {
        console.error('Andy: Web Audio API is not supported in this browser');
      }
    } catch (e) {
      console.error('Andy: Error creating Web Audio context', e);
    }
  }

  public getAnalyser(): AnalyserNode {
    return SoundSystem.analyser;
  }

  public volume(): number {
    return SoundSystem.masterGainNode.gain.value;
  }

  public adjustVolume(by: number) {
    SoundSystem.masterGainNode.gain.value = by;
  }

  public play(sound: Sound, multiple = false, loop = false, time: number = 0) {
    if (multiple === true || !sound.isPlaying) {
      sound.loop = loop;
      this.playSound(sound, time);
    }
  }

  public stop(sound: Sound) {
    if (sound.isPlaying) {
      sound.stop();
      sound.isPlaying = false;
    }
  }

  public currentTime(): number {
    if (!SoundSystem.soundContext) {
      return 0;
    }
    return SoundSystem.soundContext.currentTime;
  }

  public decodeSound(data: ArrayBuffer): Promise<AudioBuffer> {
    return new Promise<AudioBuffer>((res, rej) => {
      let buffer: AudioBuffer;
      if (SoundSystem.soundContext) {
        SoundSystem.soundContext.decodeAudioData(data, (ab: AudioBuffer) => {
          buffer = ab;
          res(buffer);
        },
          (error: any) => {
            console.error(error);
            rej(error);
          }
        );
      }
    });
  }

  public resumePlay(): Promise<void> {
    return SoundSystem.soundContext.resume();
  }

  private playSound(sound: Sound, time: number = 0) {
    sound.setContext(SoundSystem.soundContext);
    if (sound.node)
      sound.node.connect(SoundSystem.masterGainNode);
    sound.play(time);
    sound.isPlaying = true;
  }

  private handleStateChange(e: Event): any {
    if (SoundSystem.soundContext && SoundSystem.soundContext.state !== 'running') {
      this.resumePlay();
    }
    return undefined;
  }
}

//////////////////////////////////////////////////////////////////////////////

/**
 * Load a sound from the server and store it in the sound
 * registry
 * @param name
 */
export const loadSound = (url: string, soundSystem: SoundSystem): Promise<Sound> => {
  return new Promise((resolve, reject) => {
    const headers = new Headers();
    headers.append('User-Agent', 'Andy/1.0');
    // if (mimeType) {
    //   headers.append('Accept', mimeType);
    //   headers.append('Content-Type', mimeType);
    // }
    const body = undefined;
    fetch(url, {
      method: 'GET',
      mode: 'cors',
      cache: 'reload',
      headers: headers,
      credentials: 'same-origin',
      redirect: 'follow',
      referrerPolicy: 'no-referrer',
      body: (body) ? JSON.stringify(body) : body,
    })
      .then(response => response.arrayBuffer())
      .then(response => {
        soundSystem.decodeSound(<ArrayBuffer>response).then(
          (b: AudioBuffer) => {
            const sound = new Sound(b);
            sound.buffer = b;
            resolve(sound);
          },
          (err) => {
            reject(err);
          }).catch(e => console.error(e));
      }).catch(e => console.error(e));
  });
};
